import { render, html } from '/uhtml.js';
import loadThreads from '/thread.js';

function changeName(e) {
	let displayname = document.getElementById('newname').value;
	window.emit('update_user', {
		displayname
	}, msg => {
		if (!msg.success) {
			console.log('update_user error: ', msg.message);
			document.getElementById('savename').classList.remove('hidden');
			return;
		}
		window.displayname = displayname;
		document.getElementById('savename').classList.add('hidden');
	})
}

window.connectInstance = async url => {
	let instance = window.instances[url];
	let connected;
	function connecting(resolve, reject) {
		instance.socket.on('connect', resolve);
		instance.socket.on('connect_error', error => {
			console.log(`error connecting to ${url}: ${error}`);
			if (connected)
				return;
			instance.socket.disconnect();
			reject(error.message);
		});
	}
	if (instance) {
		if (!instance.socket?.connected) {
			if (instance.socket)
				instance.socket.connect();
			else
				instance.socket = io(url);
			await new Promise(connecting);
		}
	}
	else {
		instance = window.instances[url] = {
			socket: io(url),
			url
		};
		await new Promise(connecting);
	}
	connected = true;
	return instance;
};

window.getUser = async (id, name) => {
	let instance = window.instancelist[0];
	if (id) {
		id = id.split('@');
		instance = await connectInstance(id[1]);
		id = id[0];
	}
	return new Promise((resolve, reject) =>
		instance.socket.emit('get_user', { id, name }, msg => {
			if (!msg.success)
				return reject('get_user failed: ' + msg.message);
			resolve(msg.user);
		})
	);
};

async function authenticateInstance(div, select) {
	return new Promise((resolve, reject) =>
		div.instance.socket.emit('authenticate', {
			name: window.name,
			id: window.id + '@' + location.host,
			message: window.signedMessage,
			pubkey: window.keys.armored.publicKey
		}, msg => {
			if (!msg.success) {
				reject(`${div.instance.url} authentication failed: ${msg.message}`);
				return;
			}
			div.instance.id = msg.instance.id;
			div.id = 'instance' + msg.instance.id;
			div.instance.emit = (event, data, callback) =>
				div.instance.socket.emit(event, {
					...data,
					__session: window.session,
				}, callback);
			loadThreads(div, select);
			resolve();
		}));
}

function expandInstance(event) {
	this.parentElement.parentElement.children['threads'].classList.toggle('hidden');
	this.children[0].classList.toggle('collapsed');
}

function instanceClicked(event) {}

function saveInstances() {
	localStorage.setItem('instances', JSON.stringify(
		window.instancelist.map(i => ({
			id: i.id,
			url: i.url
		}))
	));
}

async function addInstance() {
	this.textContent = this.textContent.trim().toLowerCase();
	let instancediv = this.parentElement.parentElement;
	if (!this.textContent)
		return instancediv.remove();
	let url = (/^[\w-]+:.+/.test(this.textContent) ? '' : 'https://') + this.textContent;
	if (window.instancelist.find(i => i.url === url))
		return instancediv.remove();
	let instance = await connectInstance(url);
	instancediv.instance = instance;
	await authenticateInstance(instancediv, true);
	this.contentEditable = false;
	/* expander is initially just a span with a non-space space
	   in it to make the title span style work */
	let expander = this.parentElement.children[0];
	expander.classList.add('expander');
	expander.onclick = expandInstance;
	expander.innerHTML = `<span class='arrow'>◿</span>`;
	window.instancelist.push(instance);
	saveInstances();
	this.onclick = instanceClicked;
}

// main app html
document.body.append(html.node`
	<div id='home' class='column'>
		<div id='instancelist'>
			<h3>vybe</h3>
			<p id='instances'>instances:<button onclick=${() => {
				let div = html.node`
					<div>
						<div class='instance'>
							<span> </span>
							<span onblur=${addInstance} onkeydown=${function(event) {
								if (event.key === 'Enter') {
									event.preventDefault();
									addInstance.call(this);
								}
							}} class='title' contenteditable='true'>
							</span>
						</div>
					</div>`;
				document.getElementById('instancelist').append(div);
				div.children[0].children[1].focus();
			}}>add</button></p>
		</div>
		<div id='user' onclick=${() =>
			document.getElementById('profile').classList.toggle('hidden')
		}>${window.name}</div>
	</div>
	<hr class='separator' color='#505050'>
	<!-- create thread column goes here -->
	<hr class='separator' color='#505050'>
	<div id='profile' class='column hidden'>
		<p><strong>profile</strong></p>
		<label class='heading'>display name</label>
		<input id='newname' onkeyup=${function(event) {
			if (window.displayname === this.value)
				document.getElementById('savename').classList.add('hidden');
			else if (event.key === 'Enter')
				changeName();
			else
				document.getElementById('savename').classList.remove('hidden');
		}}>
		<button class='hidden' id='savename' onclick=${changeName}>save</button>
		<label class='heading'>authentication requests</label>
		<div id='authrequests'></div>
	</div>
	<hr class='separator' color='#505050'>
`);

for (let i = 0; i < instancelist.length; ++i) {
	let instance = instancelist[i];
	let div = html.node`
		<div>
			<div class='instance'>
				<span class='expander' onclick=${expandInstance}>
					<span class='arrow'>◿</span>
				</span>
				<span class='title' onclick=${instanceClicked}>${instance.url}</span>
			</div>
		</div>`;
	div.id = 'instance' + instance.id;
	div.instance = instance;
	document.getElementById('instancelist').append(div);
	if (i === 0)
		loadThreads(div, true);
	else {
		await connectInstance(instance.url);
		authenticateInstance(div);
	}
}

document.getElementById('newname').value = window.displayname;

function authRequest(authrequest) {
	const p = html.node`
		<p>
			<button onclick=${() => {
				window.emit('authorize_key', {
					id: authrequest.id
				}, msg => {
					if (!msg.success)
						console.log('authorize_key failed: ', msg.message);
				});
				p.remove();
			}}>approve</button>
			session <strong>${authrequest.id}</strong>
		</p>`;
	document.getElementById('authrequests').append(p);
	setTimeout(() => p.remove(), Date.now() - authrequest.time + 60000 * 5);
}

window.socket.on('authrequest', authRequest);

export {
	authRequest
};
