const db = require('../db');
const authwrap = require('../authwrap');
const check_permission = require('../check_permission');

async function create_thread(msg, respond) {
	// validate inputs
	if (typeof msg.name !== 'string')
		return respond({
			success: false,
			message: 'thread name required'
		});
	if (msg.name.length > 200)
		return respond({
			success: false,
			message: 'thread name 200 chars max'
		});
	if (!Array.isArray(msg.members) || msg.members.find(m => typeof m?.id !== 'string'))
		return respond({
			success: false,
			message: 'invalid members'
		});
	if (msg.members.find(m => m.id == msg.auth_user.id)?.permissions?.admin !== 'true')
		return respond({
			success: false,
			message: 'creator missing admin permission'
		});
	// add to db
	const thread_id = (await db.query(
		'insert into thread (name, creator) values (?, ?) returning id',
		[msg.name, msg.auth_user.id]
	)).rows[0].id;
	// set up permissions
	if (!msg.permissions?.view_limited) {
		await db.query(
			`insert into permission (thread, type, mutable, permission, value) 
			values (?, ?, ?, ?, ?)`,
			[thread_id, 'everyone', true, 'view', 'true']
		);
		if (!msg.permissions?.post_limited) {
			await db.query(
				`insert into permission (thread, type, mutable, permission, value) 
				values (?, ?, ?, ?, ?)`,
				[thread_id, 'everyone', true, 'post', 'true']
			);
		} else {
			await db.query(
				`insert into permission (thread, type, mutable, permission, value) 
				values (?, ?, ?, ?, ?)`,
				[thread_id, 'members', true, 'post', 'true']
			);
		}
	} else {
		await db.query(
			`insert into permission (thread, type, mutable, permission, value) 
			values (?, ?, ?, ?, ?)`,
			[thread_id, 'members', true, 'view', 'true']
		);
		await db.query(
			`insert into permission (thread, type, mutable, permission, value) 
			values (?, ?, ?, ?, ?)`,
			[thread_id, 'members', true, 'post', 'true']
		);
	}
	// add members
	let members = {};
	for (let member of msg.members) {
		let id = member.id.split('@');
		id = id[1] === vybe.instance.url ? id[0] : member.id;
		if (members[id])
			continue;
		await db.query(
			'insert into member (thread, user) values (?, ?)',
			[thread_id, id]
		);
		members[id] = true;
		if (typeof member.permissions === 'object')
			for (let permission in member.permissions)
				await db.query(`
					insert into permission (thread, type, user, mutable, permission, value)
					values (?, ?, ?, ?, ?, ?)`,
					[thread_id, 'user', id, true, permission, member.permissions[permission]]);
	}
	let thread = {
		id: thread_id,
		name: msg.name,
		streams: []
	};
	if (!msg.permissions?.view_limited) {
		for (let id in vybe.users) {
			for (let socket of vybe.users[id].sockets) {
				socket.emit('thread', {
					...thread,
					permissions: {
						view: true,
						post: !msg.permissions?.post_limited || members[id],
						admin: id === msg.auth_user.id
					}
				});
			}
		}
	}
	else {
		for (let member in members) {
			member = vybe.users[member];
			if (!member)
				continue;
			for (let socket of member.sockets) {
				socket.emit('thread', {
					...thread,
					permissions: {
						view: true,
						post: true,
						admin: member.id === msg.auth_user.id
					}
				});
			}
		}
	}
	// respond
	return respond({
		success: true,
		id: thread_id
	});
}

async function list_threads(msg, respond) {
	// this can be optimized by merging the permission check into here
	let threads = await db.query(
		`select name, id from thread
		join permission on thread.id = permission.thread
		left join member on thread.id = member.thread
		where permission.permission = 'view' and permission.value = 'true'
		and ((permission.type = 'everyone') or
			permission.type = 'members' and member.user = ?)
		group by thread.id
		order by thread.created desc`,
		msg.auth_user.id
	);
	threads = await Promise.all(threads.rows.map(async thread => {
		thread.streams = vybe.threads[thread.id]?.streams || [];
		thread.permissions = await check_permission(msg.auth_user.id, thread.id);
		return thread;
	}));
	return respond({
		success: true,
		threads
	});
}

async function get_thread(msg, respond) {
	if (!msg.thread) {
		return respond({
			success: false,
			message: 'thread ID required'
		});
	}
	let perms = await check_permission(msg.auth_user.id, msg.thread);
	if (!perms.view) {
		return respond({
			success: false,
			message: "you can't view this thread"
		});
	}
	let thread = await db.query(
		`select thread.name, member.user, user.name as username,
			coalesce(user.displayname, user.name) as displayname
		from thread
		left join member on thread.id = member.thread
		left join user on user.id = member.user
		where thread.id = ?`,
		msg.thread
	);
	const permissions = await db.query(
		`select type, user, permission, value, mutable
		from permission where thread = ?`,
		msg.thread
	);
	let members = Object.fromEntries(thread.rows.map(member =>
		[member.user, {
			id: member.user,
			name: member.username,
			displayname: member.displayname
		}]
	));
	for (let permission of permissions.rows) {
		const member = members[permission.user];
		if (member)
			(member.permissions || (member.permissions = {}))
				[permission.permission] = {
					value: permission.value,
					mutable: permission.mutable
				};
		else
			(perms[permission.type] || (perms[permission.type] = {}))
				[permission.permission] = {
					value: permission.value,
					mutable: permission.mutable
				};
	}
	function makeBool(type, permission, fallback) {
		if (perms[type]) {
			if (perms[type][permission])
				perms[type][permission].value = perms[type][permission].value === 'true';
			else
				perms[type][permission] = { value: fallback, mutable: true };
		} else
			(perms[type] = {})[permission] = { value: fallback, mutable: true };
	}
	makeBool('everyone', 'view', false);
	makeBool('everyone', 'post', false);
	makeBool('members', 'view', true);
	makeBool('members', 'post', true);
	thread = {
		id: msg.thread,
		name: thread.rows[0].name,
		permissions: perms,
		members: Object.values(members),
		streams: vybe.threads[msg.thread]?.streams || []
	};
	return respond({
		success: true,
		thread
	});
}

async function edit_thread(msg, respond) {
	// validate inputs
	if (!msg.id || typeof msg.name !== 'string') {
		return respond({
			success: false,
			message: 'invalid msg'
		});
	}
	if (msg.name.length > 200) {
		return respond({
			success: false,
			message: 'thread name 200 chars max'
		});
	}
	const perms = await check_permission(msg.auth_user.id, msg.id);
	if (!perms.admin) {
		return respond({
			success: false,
			message: "user doesn't have permission"
		});
	}
	// update name
	await db.query(
		'update thread set name = ? where id = ?',
		[msg.name, msg.id]
	);
	// update permissions
	let permissions = {};
	for (const p of (await db.query(
		`select type, permission, value, mutable
		from permission
		where type != 'user' and thread = ?`,
		msg.id
	)).rows) {
		(permissions[p.type] || (permissions[p.type] = {}))
			[p.permission] = {
				value: p.value,
				mutable: p.mutable
			};
	}
	async function setPermission(type, permission, value) {
		if (permissions[type] && permissions[type][permission]) {
			if (!permissions[type][permission].mutable) {
				respond({
					success: false,
					message: 'permission not mutable'
				});
				return false;
			}
			if (permissions[type][permission].value !== value) {
				await db.query(`
					update permission set value = ?
					where thread = ? and type = ? and permission = ?`,
					[value, msg.id, type, permission]
				);
				permissions[type][permission].value = value;
			}
		}
		else {
			await db.query(
				`insert into permission (thread, type, mutable, permission, value) 
				values (?, ?, ?, ?, ?)`,
				[msg.id, type, true, permission, value]
			);
			(permissions[type] || (permissions[type] = {}))
				[permission] = {
					value,
					mutable: true
				};
		}
		return true;
	}
	if (!msg.permissions || !msg.permissions.view_limited) {
		if (!await setPermission('everyone', 'view', 'true'))
			return;
		if (!msg.permissions || !msg.permissions.post_limited) {
			if (!await setPermission('everyone', 'post', 'true'))
				return;
		}
		else {
			if (!await setPermission('members', 'post', 'true'))
				return;
			if (!await setPermission('everyone', 'post', 'false'))
				return;
		}
	} else {
		if (!await setPermission('members', 'view', 'true'))
			return;
		if (!await setPermission('members', 'post', 'true'))
			return;
		if (!await setPermission('everyone', 'view', 'false'))
			return;
		if (!await setPermission('everyone', 'post', 'false'))
			return;
	}
	permissions.everyone.view.value = permissions.everyone.view.value === 'true';
	permissions.everyone.post.value = permissions.everyone.post.value === 'true';
	if (!permissions.members) permissions.members = {};
	if (!permissions.members.view)
		permissions.members.view = { value: 'true', mutable: true };
	if (!permissions.members.post)
		permissions.members.post = { value: 'true', mutable: true };
	permissions.members.view.value = permissions.members.view.value === 'true';
	permissions.members.post.value = permissions.members.post.value === 'true';
	let members = Object.fromEntries((await db.query(
			`select member.user from thread
			join member on thread.id = member.thread
			where thread.id = ?`,
			msg.id
		)).rows.map(row => [row.user, true]));
	let thread = {
		id: msg.id,
		name: msg.name,
		streams: vybe.threads[msg.id]?.streams || []
	};
	if (!msg.permissions?.view_limited) {
		for (let id in vybe.users)
			for (let socket of vybe.users[id].sockets)
				socket.emit('thread', {
					...thread,
					permissions: {
						view: true,
						post: !msg.permissions?.post_limited || id in members,
						admin: id === msg.auth_user.id && perms.admin,
						...permissions
					}
				});
	}
	else {
		for (let member in members) {
			member = vybe.users[member];
			if (!member)
				continue;
			for (let socket of member.sockets)
				socket.emit('thread', {
					...thread,
					permissions: {
						view: true,
						post: true,
						admin: member.id === msg.auth_user.id && perms.admin,
						...permissions
					}
				});
		}
	}
	return respond({
		success: true
	});
}

module.exports = {
	create_thread: authwrap(create_thread),
	list_threads: authwrap(list_threads),
	get_thread: authwrap(get_thread),
	edit_thread: authwrap(edit_thread)
};
